@@ -50,14 +50,15 @@ var Cc = Components.classes;
 var Cr = Components.results;
 
 const POST_DATA_ANNO = "bookmarkProperties/POSTData";
+const READ_ONLY_ANNO = "placesInternal/READ_ONLY";
 const LMANNO_FEEDURI = "livemark/feedURI";
 const LMANNO_SITEURI = "livemark/siteURI";
 
-//@line 57 "/builds/tinderbox/XR-Trunk/Darwin_8.8.4_Depend/mozilla/toolkit/components/places/src/utils.js"
+//@line 58 "/builds/tinderbox/Xr-Mozilla1.9-Release/Darwin_8.8.4_Depend/mozilla/toolkit/components/places/src/utils.js"
 // On Mac OSX, the transferable system converts "\r\n" to "\n\n", where we
 // really just want "\n".
 const NEWLINE= "\n";
-//@line 64 "/builds/tinderbox/XR-Trunk/Darwin_8.8.4_Depend/mozilla/toolkit/components/places/src/utils.js"
+//@line 65 "/builds/tinderbox/Xr-Mozilla1.9-Release/Darwin_8.8.4_Depend/mozilla/toolkit/components/places/src/utils.js"
 
 function QI_node(aNode, aIID) {
   var result = null;
@@ -252,8 +253,10 @@ var PlacesUtils = {
   nodeIsReadOnly: function PU_nodeIsReadOnly(aNode) {
     if (this.nodeIsFolder(aNode))
       return this.bookmarks.getFolderReadonly(asQuery(aNode).folderItemId);
-    if (this.nodeIsQuery(aNode))
-      return asQuery(aNode).childrenReadOnly;
+    if (this.nodeIsQuery(aNode) &&
+        asQuery(aNode).queryOptions.resultType !=
+          Ci.nsINavHistoryQueryOptions.RESULTS_AS_TAG_CONTENTS)
+      return aNode.childrenReadOnly;
     return false;
   },
 
@@ -285,6 +288,18 @@ var PlacesUtils = {
              resultType == Ci.nsINavHistoryQueryOptions.RESULTS_AS_DATE_SITE_QUERY);
   },
 
+  /**
+   * Determines whether or not a result-node is a tag container.
+   * @param   aNode
+   *          A result-node
+   * @returns true if the node is a tag container, false otherwise
+   */
+  nodeIsTagQuery: function PU_nodeIsTagQuery(aNode) {
+    return aNode.type == Ci.nsINavHistoryResultNode.RESULT_TYPE_QUERY &&
+           asQuery(aNode).queryOptions.resultType ==
+             Ci.nsINavHistoryQueryOptions.RESULTS_AS_TAG_CONTENTS;
+  },
+
   /**
    * Determines whether or not a ResultNode is a container.
    * @param   aNode
@@ -373,6 +388,13 @@ var PlacesUtils = {
   getConcreteItemId: function PU_getConcreteItemId(aNode) {
     if (aNode.type == Ci.nsINavHistoryResultNode.RESULT_TYPE_FOLDER_SHORTCUT)
       return asQuery(aNode).folderItemId;
+    else if (PlacesUtils.nodeIsTagQuery(aNode)) {
+      // RESULTS_AS_TAG_CONTENTS queries are similar to folder shortcuts
+      // so we can still get the concrete itemId for them.
+      var queries = aNode.getQueries({});
+      var folders = queries[0].getFolders({});
+      return folders[0];
+    }
     return aNode.itemId;
   },
 
@@ -415,9 +437,11 @@ var PlacesUtils = {
    *          Used instead of the node's URI if provided.
    *          This is useful for wrapping a container as TYPE_X_MOZ_URL,
    *          TYPE_HTML or TYPE_UNICODE.
+   * @param   aForceCopy
+   *          Does a full copy, resolving folder shortcuts.
    * @returns A string serialization of the node
    */
-  wrapNode: function PU_wrapNode(aNode, aType, aOverrideURI) {
+  wrapNode: function PU_wrapNode(aNode, aType, aOverrideURI, aForceCopy) {
     var self = this;
 
     // when wrapping a node, we want all the items, even if the original
@@ -425,8 +449,10 @@ var PlacesUtils = {
     // this can happen when copying from the left hand pane of the bookmarks
     // organizer
     function convertNode(cNode) {
-      if (self.nodeIsFolder(cNode) && asQuery(cNode).queryOptions.excludeItems)
-        return self.getFolderContents(cNode.itemId, false, true).root;
+      if (self.nodeIsFolder(cNode) && asQuery(cNode).queryOptions.excludeItems) {
+        var concreteId = self.getConcreteItemId(cNode);
+        return self.getFolderContents(concreteId, false, true).root;
+      }
       return cNode;
     }
 
@@ -440,7 +466,7 @@ var PlacesUtils = {
             this.value += aStr;
           }
         };
-        self.serializeNodeAsJSONToOutputStream(convertNode(aNode), writer, true);
+        self.serializeNodeAsJSONToOutputStream(convertNode(aNode), writer, true, aForceCopy);
         return writer.value;
       case this.TYPE_X_MOZ_URL:
         function gatherDataUrl(bNode) {
@@ -466,7 +492,7 @@ var PlacesUtils = {
             return s;
           }
           // escape out potential HTML in the title
-          var escapedTitle = htmlEscape(bNode.title);
+          var escapedTitle = bNode.title ? htmlEscape(bNode.title) : "";
           if (self.nodeIsLivemarkContainer(bNode)) {
             var siteURI = self.livemarks.getSiteURI(bNode.itemId).spec;
             return "<A HREF=\"" + siteURI + "\">" + escapedTitle + "</A>" + NEWLINE;
@@ -568,7 +594,8 @@ var PlacesUtils = {
           // note:  this._uri() will throw if uriString is not a valid URI
           if (this._uri(uriString)) {
             nodes.push({ uri: uriString,
-                         title: titleString ? titleString : uriString });
+                         title: titleString ? titleString : uriString ,
+                         type: this.TYPE_X_MOZ_URL });
           }
         }
         break;
@@ -578,7 +605,9 @@ var PlacesUtils = {
           var uriString = parts[i];
           // note: this._uri() will throw if uriString is not a valid URI
           if (uriString != "" && this._uri(uriString))
-            nodes.push({ uri: uriString, title: uriString });
+            nodes.push({ uri: uriString,
+                         title: uriString,
+                         type: this.TYPE_X_MOZ_URL });
         }
         break;
       default:
@@ -966,9 +995,15 @@ var PlacesUtils = {
    * Restores bookmarks/tags from a JSON file.
    * WARNING: This method *removes* any bookmarks in the collection before
    * restoring from the file.
+   *
+   * @param aFile
+   *        nsIFile of bookmarks in JSON format to be restored.
+   * @param aExcludeItems
+   *        Array of root item ids (ie: children of the places root)
+   *        to not delete when restoring.
    */
   restoreBookmarksFromJSONFile:
-  function PU_restoreBookmarksFromJSONFile(aFile) {
+  function PU_restoreBookmarksFromJSONFile(aFile, aExcludeItems) {
     // open file stream
     var stream = Cc["@mozilla.org/network/file-input-stream;1"].
                  createInstance(Ci.nsIFileInputStream);
@@ -988,7 +1023,7 @@ var PlacesUtils = {
     if (jsonStr.length == 0)
       return; // empty file
 
-    this.restoreBookmarksFromJSONString(jsonStr, true);
+    this.restoreBookmarksFromJSONString(jsonStr, true, aExcludeItems);
   },
 
   /**
@@ -998,9 +1033,12 @@ var PlacesUtils = {
    *        JSON string of serialized bookmark data.
    * @param aReplace
    *        Boolean if true, replace existing bookmarks, else merge.
+   * @param aExcludeItems
+   *        Array of root item ids (ie: children of the places root)
+   *        to not delete when restoring.
    */
   restoreBookmarksFromJSONString:
-  function PU_restoreBookmarksFromJSONString(aString, aReplace) {
+  function PU_restoreBookmarksFromJSONString(aString, aReplace, aExcludeItems) {
     // convert string to JSON
     var nodes = this.unwrapNodes(aString, this.TYPE_X_MOZ_PLACE_CONTAINER);
 
@@ -1015,58 +1053,112 @@ var PlacesUtils = {
     });
 
     var batch = {
+      _utils: this,
       nodes: nodes[0].children,
       runBatched: function restore_runBatched() {
-        this.nodes.forEach(function(node) {
-          var root = node.root;
-          // FIXME support folders other than known roots
-          // restoring the library left pane, for example, breaks the library
-          if (!root)
-            return;
+        if (aReplace) {
+          var excludeItems = aExcludeItems || [];
+          // delete existing children of the root node, excepting:
+          // 1. special folders: delete the child nodes 
+          // 2. tags folder: untag via the tagging api
+          var query = this._utils.history.getNewQuery();
+          query.setFolders([this._utils.placesRootId], 1);
+          var options = this._utils.history.getNewQueryOptions();
+          options.expandQueries = false;
+          var root = this._utils.history.executeQuery(query, options).root;
+          root.containerOpen = true;
+          var childIds = [];
+          for (var i = 0; i < root.childCount; i++) {
+            var childId = root.getChild(i).itemId;
+            if (excludeItems.indexOf(childId) == -1)
+              childIds.push(childId);
+          }
+          root.containerOpen = false;
 
+          for (var i = 0; i < childIds.length; i++) {
+            var rootItemId = childIds[i];
+            if (rootItemId == this._utils.tagsFolderId) {
+              // remove tags via the tagging service
+              var tags = this._utils.tagging.allTags;
+              var uris = [];
+              var bogusTagContainer = false;
+              for (let i in tags) {
+                var tagURIs = [];
+                // skip empty tags since getURIsForTag would throw
+                if (tags[i])
+                  tagURIs = this._utils.tagging.getURIsForTag(tags[i]);
+
+                if (!tagURIs.length) {
+                  // This is a bogus tag container, empty tags should be removed
+                  // automatically, but this does not work if they contain some
+                  // not-uri node, so we remove them manually.
+                  // XXX this is a temporary workaround until we implement
+                  // preventive database maintainance in bug 431558.
+                  bogusTagContainer = true;
+                }
+                for (let j in tagURIs)
+                  this._utils.tagging.untagURI(tagURIs[j], [tags[i]]);
+              }
+              if (bogusTagContainer)
+                this._utils.bookmarks.removeFolderChildren(rootItemId);
+            }
+            else if ([this._utils.toolbarFolderId,
+                      this._utils.unfiledBookmarksFolderId,
+                      this._utils.bookmarksMenuFolderId].indexOf(rootItemId) != -1)
+              this._utils.bookmarks.removeFolderChildren(rootItemId);
+            else
+              this._utils.bookmarks.removeItem(rootItemId);
+          }
+        }
+
+        var searchIds = [];
+        var folderIdMap = [];
+
+        this.nodes.forEach(function(node) {
           if (!node.children || node.children.length == 0)
             return; // nothing to restore for this root
 
-          var container = this.placesRootId; // default to places root
-          switch (root) {
-            case "bookmarksMenuFolder":
-              container = this.bookmarksMenuFolderId;
-              break;
-            case "tagsFolder":
-              container = this.tagsFolderId;
-              break;
-            case "unfiledBookmarksFolder":
-              container = this.unfiledBookmarksFolderId;
-              break;
-            case "toolbarFolder":
-              container = this.toolbarFolderId;
-              break;
-          }
-
-          if (aReplace) {
-            if (container != this.tagsFolderId)
-              this.bookmarks.removeFolderChildren(container);
-            else {
-                // remove tags via the tagging service
-                var tags = this.tagging.allTags;
-                var uris = [];
-                tags.forEach(function(aTag) {
-                  var tagURIs = this.tagging.getURIsForTag(aTag);
-                  for (let i in tagURIs)
-                    this.tagging.untagURI(tagURIs[i], [aTag]);
-                }, this);
+          if (node.root) {
+            var container = this.placesRootId; // default to places root
+            switch (node.root) {
+              case "bookmarksMenuFolder":
+                container = this.bookmarksMenuFolderId;
+                break;
+              case "tagsFolder":
+                container = this.tagsFolderId;
+                break;
+              case "unfiledBookmarksFolder":
+                container = this.unfiledBookmarksFolderId;
+                break;
+              case "toolbarFolder":
+                container = this.toolbarFolderId;
+                break;
             }
+ 
+            // insert the data into the db
+            node.children.forEach(function(child) {
+              var index = child.index;
+              var [folders, searches] = this.importJSONNode(child, container, index);
+              folderIdMap = folderIdMap.concat(folders);
+              searchIds = searchIds.concat(searches);
+            }, this);
           }
+          else
+            this.importJSONNode(node, this.placesRootId, node.index);
 
-          // insert the data into the db
-          node.children.forEach(function(child) {
-            var index = child.index;
-            this.importJSONNode(child, container, index);
-          }, this);
-        }, PlacesUtils);
+        }, this._utils);
+
+        // fixup imported place: uris that contain folders
+        searchIds.forEach(function(aId) {
+          var oldURI = this.bookmarks.getBookmarkURI(aId);
+          var uri = this._fixupQuery(this.bookmarks.getBookmarkURI(aId),
+                                     folderIdMap);
+          if (!uri.equals(oldURI))
+            this.bookmarks.changeBookmarkURI(aId, uri);
+        }, this._utils);
       }
     };
-    
+
     this.bookmarks.runInBatchMode(batch, null);
   },
 
@@ -1079,18 +1171,26 @@ var PlacesUtils = {
    *          The container the data was dropped or pasted into
    * @param   aIndex
    *          The index within the container the item was dropped or pasted at
+   * @returns an array containing of maps of old folder ids to new folder ids,
+   *          and an array of saved search ids that need to be fixed up.
+   *          eg: [[[oldFolder1, newFolder1]], [search1]]
    */
   importJSONNode: function PU_importJSONNode(aData, aContainer, aIndex) {
-    // create item
+    var folderIdMap = [];
+    var searchIds = [];
     var id = -1;
     switch (aData.type) {
       case this.TYPE_X_MOZ_PLACE_CONTAINER:
         if (aContainer == PlacesUtils.bookmarks.tagsFolder) {
           if (aData.children) {
             aData.children.forEach(function(aChild) {
-              this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              try {
+                this.tagging.tagURI(this._uri(aChild.uri), [aData.title]);
+              } catch (ex) {
+                // invalid tag child, skip it
+              }
             }, this);
-            return;
+            return [folderIdMap, searchIds];
           }
         }
         else if (aData.livemark && aData.annos) {
@@ -1114,10 +1214,13 @@ var PlacesUtils = {
         }
         else {
           id = this.bookmarks.createFolder(aContainer, aData.title, aIndex);
+          folderIdMap.push([aData.id, id]);
           // process children
           if (aData.children) {
             aData.children.every(function(aChild, aIndex) {
-              this.importJSONNode(aChild, id, aIndex);
+              var [folderIds, searches] = this.importJSONNode(aChild, id, aIndex);
+              folderIdMap = folderIdMap.concat(folderIds);
+              searchIds = searchIds.concat(searches);
               return true;
             }, this);
           }
@@ -1134,6 +1237,8 @@ var PlacesUtils = {
         }
         if (aData.charset)
           this.history.setCharsetForURI(this._uri(aData.uri), aData.charset);
+        if (aData.uri.match(/^place:/))
+          searchIds.push(id);
         break;
       case this.TYPE_X_MOZ_PLACE_SEPARATOR:
         id = this.bookmarks.insertSeparator(aContainer, aIndex);
@@ -1148,6 +1253,45 @@ var PlacesUtils = {
       if (aData.annos)
         this.setAnnotationsForItem(id, aData.annos);
     }
+
+    return [folderIdMap, searchIds];
+  },
+
+  /**
+   * Replaces imported folder ids with their local counterparts in a place: URI.
+   *
+   * @param   aURI
+   *          A place: URI with folder ids.
+   * @param   aFolderIdMap
+   *          An array mapping old folder id to new folder ids.
+   * @returns the fixed up URI if all matched. If some matched, it returns
+   *          the URI with only the matching folders included. If none matched it
+   *          returns the input URI unchanged.
+   */
+  _fixupQuery: function PU__fixupQuery(aQueryURI, aFolderIdMap) {
+    var queries = {};
+    var options = {};
+    this.history.queryStringToQueries(aQueryURI.spec, queries, {}, options);
+
+    var fixedQueries = [];
+    queries.value.forEach(function(aQuery) {
+      var folders = aQuery.getFolders({});
+
+      var newFolders = [];
+      aFolderIdMap.forEach(function(aMapping) {
+        if (folders.indexOf(aMapping[0]) != -1)
+          newFolders.push(aMapping[1]);
+      });
+
+      if (newFolders.length)
+        aQuery.setFolders(newFolders, newFolders.length);
+      fixedQueries.push(aQuery);
+    });
+
+    var stringURI = this.history.queriesToQueryString(fixedQueries,
+                                                      fixedQueries.length,
+                                                      options.value);
+    return this._uri(stringURI);
   },
 
   /**
@@ -1163,9 +1307,15 @@ var PlacesUtils = {
    * @param   aIsUICommand
    *          Boolean - If true, modifies serialization so that each node self-contained.
    *          For Example, tags are serialized inline with each bookmark.
+   * @param   aResolveShortcuts
+   *          Converts folder shortcuts into actual folders. 
+   * @param   aExcludeItems
+   *          An array of item ids that should not be written to the backup.
    */
   serializeNodeAsJSONToOutputStream:
-  function PU_serializeNodeAsJSONToOutputStream(aNode, aStream, aIsUICommand) {
+  function PU_serializeNodeAsJSONToOutputStream(aNode, aStream, aIsUICommand,
+                                                aResolveShortcuts,
+                                                aExcludeItems) {
     var self = this;
     
     function addGenericProperties(aPlacesNode, aJSNode) {
@@ -1192,9 +1342,13 @@ var PlacesUtils = {
             // backup/restore of non-whitelisted annos
             // XXX causes JSON encoding errors, so utf-8 encode
             //anno.value = unescape(encodeURIComponent(anno.value));
-            if (anno.name == "livemark/feedURI")
+            if (anno.name == LMANNO_FEEDURI)
               aJSNode.livemark = 1;
-            return anno.name != "placesInternal/GUID";
+            if (anno.name == READ_ONLY_ANNO && aResolveShortcuts) {
+              // When copying a read-only node, remove the read-only annotation.
+              return false;
+            }
+            return true;
           });
         } catch(ex) {
           LOG(ex);
@@ -1232,14 +1386,19 @@ var PlacesUtils = {
 
     function addContainerProperties(aPlacesNode, aJSNode) {
       // saved queries
-      if (aJSNode.id != -1 &&
-          self.bookmarks.getItemType(aJSNode.id) == self.bookmarks.TYPE_BOOKMARK) {
+      var concreteId = PlacesUtils.getConcreteItemId(aPlacesNode);
+      if (aJSNode.id != -1 && (PlacesUtils.nodeIsQuery(aPlacesNode) ||
+          (concreteId != aPlacesNode.itemId && !aResolveShortcuts))) {
         aJSNode.type = self.TYPE_X_MOZ_PLACE;
         aJSNode.uri = aPlacesNode.uri;
-        aJSNode.concreteId = PlacesUtils.getConcreteItemId(aPlacesNode);
+        // folder shortcut
+        if (aIsUICommand)
+          aJSNode.concreteId = concreteId;
         return;
       }
       else if (aJSNode.id != -1) { // bookmark folder
+        if (concreteId != aPlacesNode.itemId)
+        aJSNode.type = self.TYPE_X_MOZ_PLACE;
         aJSNode.type = self.TYPE_X_MOZ_PLACE_CONTAINER;
         // mark special folders
         if (aJSNode.id == self.bookmarks.placesRoot)
@@ -1279,14 +1438,20 @@ var PlacesUtils = {
       // write child nodes
       if (!aNode.livemark) {
         asContainer(aSourceNode);
-        aSourceNode.containerOpen = true;
+        var wasOpen = aSourceNode.containerOpen;
+        if (!wasOpen)
+          aSourceNode.containerOpen = true;
         var cc = aSourceNode.childCount;
         for (var i = 0; i < cc; ++i) {
-          if (i != 0)
+          var childNode = aSourceNode.getChild(i);
+          if (aExcludeItems && aExcludeItems.indexOf(childNode.itemId) != -1)
+            continue;
+          var written = serializeNodeToJSONStream(aSourceNode.getChild(i), i);
+          if (written && i < cc - 1)
             aStream.write(",", 1);
-          serializeNodeToJSONStream(aSourceNode.getChild(i), i);
         }
-        aSourceNode.containerOpen = false;
+        if (!wasOpen)
+          aSourceNode.containerOpen = false;
       }
 
       // write suffix
@@ -1304,24 +1469,53 @@ var PlacesUtils = {
 
       addGenericProperties(bNode, node);
 
-      if (self.nodeIsURI(bNode))
+      var parent = bNode.parent;
+      var grandParent = parent ? parent.parent : null;
+
+      if (self.nodeIsURI(bNode)) {
+        // Tag root accept only folder nodes
+        if (parent && parent.itemId == self.tagsFolderId)
+          return false;
+        // Check for url validity, since we can't halt while writing a backup.
+        // This will throw if we try to serialize an invalid url and it does
+        // not make sense saving a wrong or corrupt uri node.
+        try {
+          self._uri(bNode.uri);
+        } catch (ex) {
+          return false;
+        }
         addURIProperties(bNode, node);
-      else if (self.nodeIsContainer(bNode))
+      }
+      else if (self.nodeIsContainer(bNode)) {
+        // Tag containers accept only uri nodes
+        if (grandParent && grandParent.itemId == self.tagsFolderId)
+          return false;
         addContainerProperties(bNode, node);
-      else if (self.nodeIsSeparator(bNode))
+      }
+      else if (self.nodeIsSeparator(bNode)) {
+        // Tag root accept only folder nodes
+        // Tag containers accept only uri nodes
+        if ((parent && parent.itemId == self.tagsFolderId) ||
+            (grandParent && grandParent.itemId == self.tagsFolderId))
+          return false;
+
         addSeparatorProperties(bNode, node);
+      }
 
       if (!node.feedURI && node.type == self.TYPE_X_MOZ_PLACE_CONTAINER)
         writeComplexNode(aStream, node, bNode);
       else
         writeScalarNode(aStream, node);
+      return true;
     }
 
     // serialize to stream
     serializeNodeToJSONStream(aNode, null);
   },
 
-  // XXX testing serializers
+  /**
+   * Serialize a JS object to JSON
+   */
   toJSONString: function PU_toJSONString(aObj) {
     var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);
     return JSON.encode(aObj);
@@ -1330,7 +1524,7 @@ var PlacesUtils = {
   /**
    * Serializes bookmarks using JSON, and writes to the supplied file.
    */
-  backupBookmarksToFile: function PU_backupBookmarksToFile(aFile) {
+  backupBookmarksToFile: function PU_backupBookmarksToFile(aFile, aExcludeItems) {
     if (aFile.exists() && !aFile.isWritable())
       return; // XXX
 
@@ -1360,7 +1554,8 @@ var PlacesUtils = {
     var result = this.history.executeQuery(query, options);
     result.root.containerOpen = true;
     // serialize as JSON, write to stream
-    this.serializeNodeAsJSONToOutputStream(result.root, streamProxy);
+    this.serializeNodeAsJSONToOutputStream(result.root, streamProxy,
+                                           false, false, aExcludeItems);
     result.root.containerOpen = false;
 
     // close converter and stream
